using System;
using System.Runtime.InteropServices;

namespace GpioDotNet
{

    class GpioCmT35 : IGpioInterface
    {
        private IntPtr hDrv = IntPtr.Zero;
        

        // Device CreateFile specific constants
        internal const Int32 INVALID_HANDLE_VALUE = -1;
       
        internal const uint GPIO_DIR_INPUT    = (1 << 0);
        internal const uint GPIO_DIR_OUTPUT   = (0 << 0);
        internal const uint GPIO_INT_LOW_HIGH = (1 << 1);
        internal const uint GPIO_INT_HIGH_LOW = (1 << 2);
        internal const uint GPIO_INT_LOW      = (1 << 3);
        internal const uint GPIO_INT_HIGH     = (1 << 4);

        internal uint IOCTL_GPIO_SET_BIT = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0300);
        internal uint IOCTL_GPIO_CLR_BIT = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0301);
        internal uint IOCTL_GPIO_GETBIT  = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0303);
        internal uint IOCTL_GPIO_SETMODE = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0304);
        internal uint IOCTL_GPIO_GETMODE = NativeMethods.CTL_CODE_DEVICE_UNKNOWN(0x0305);

 
        [StructLayout(LayoutKind.Sequential)]
        internal class IoctlGpioInfo
        {
            public IoctlGpioInfo(uint gpio, uint direction)
            {
                this.num = gpio;
                this.dir = direction;
            }
            public uint num;
            public uint dir;
        }

        [StructLayout(LayoutKind.Sequential)]
        internal class IoctlGpioInitInterruptInfo
        {
            public IoctlGpioInitInterruptInfo(uint num, IntPtr eventHandle)
            {
                this.uiGpioID = num;
                //this.uiSysIntrID = irq;
                this.hEvent =  (uint)eventHandle.ToInt32();
            }
            public uint uiGpioID;
            //public uint uiSysIntrID;
            public uint hEvent;
        }
        [StructLayout(LayoutKind.Sequential)]
        internal class IoctlGpioInterruptInfo
        {
            public IoctlGpioInterruptInfo(uint num, uint irq)
            {
                this.uiGpioID = num;
                this.uiSysIntrID = irq;
            }
            public uint uiGpioID;
            public uint uiSysIntrID;
        }

        private bool SendIOCTL(uint ioctl, Object inputData, uint inputParametersSize)
        {
            UInt32 stam;
            return SendIOCTL(ioctl, inputData, inputParametersSize, out stam, 0);
        }

        private UInt32 GetIOCTL(uint ioctl, uint value)
        {
            UInt32 gpioNumber = value;
            UInt32 retVal = 0;

            SendIOCTL(ioctl, gpioNumber, (uint)Marshal.SizeOf(gpioNumber), out retVal, (uint)Marshal.SizeOf(retVal));
            return retVal;
        }

        private bool SendIOCTL(uint ioctl, Object inputData, uint inputParametersSize, out UInt32 outputData, uint outputParametersSize)
        {
            uint bytesReturned = 0;
            // Initialize unmanged memory to hold the struct.
            IntPtr inputParametersPtr = IntPtr.Zero;
            IntPtr outputParametersPtr = IntPtr.Zero;
            try
            {
                if (inputParametersSize > 0)
                {
                    inputParametersPtr = Marshal.AllocHGlobal((int)inputParametersSize);
                    if (inputParametersSize > 4)
                        Marshal.StructureToPtr(inputData, inputParametersPtr, false);
                    else
                        Marshal.WriteInt32(inputParametersPtr,Convert.ToInt32(inputData));
                            
                }
                if(outputParametersSize > 0)
                    outputParametersPtr = Marshal.AllocHGlobal((int)outputParametersSize);
                // Copy the struct to unmanaged memory.
                
                bool res = NativeMethods.DeviceIoControl(hDrv, ioctl,
                                           inputParametersPtr, inputParametersSize,
                                           outputParametersPtr, outputParametersSize,
                                           ref bytesReturned, IntPtr.Zero);
                if (!res)
                {
                    // Throw Win32 exception with error code
                    int err = Marshal.GetLastWin32Error();
                    throw new System.ComponentModel.Win32Exception(err);
                }
                if (outputParametersSize > 0)
                {
                    outputData = (UInt32)Marshal.ReadInt32(outputParametersPtr);
                }
                else
                    outputData = 0;
            }
            catch (Exception ex)
            {
                // throw all other exceptions to upper levels
                throw ex;
            }
            finally
            {
                if(inputParametersSize > 0)
                    Marshal.FreeHGlobal(inputParametersPtr);
                if(outputParametersSize > 0)
                    Marshal.FreeHGlobal(outputParametersPtr);
            }
            return true;
        }

        
        public GpioCmT35() 
        {
            //First, we create a handle to the I2C proxy driver. I2C3 in our case.
            hDrv = NativeMethods.CreateFile("GIO1:", 0, 0, IntPtr.Zero, 0, 0, IntPtr.Zero);
            if (hDrv.ToInt32() == INVALID_HANDLE_VALUE)
                throw new InvalidOperationException("Can't open GIO1:");
        }

        ~GpioCmT35()
        {
            NativeMethods.CloseHandle(hDrv);
        }

        public bool SetAsOutput(uint gpioNumber, GpioLevel level)
        {
            IoctlGpioInfo inputParameters = new IoctlGpioInfo(gpioNumber, GPIO_DIR_OUTPUT);
            
            bool res = SendIOCTL(IOCTL_GPIO_SETMODE, inputParameters,(uint) Marshal.SizeOf(inputParameters));
            
            if (!res)
                return false;

            if (level == GpioLevel.HIGH)
                return SendIOCTL(IOCTL_GPIO_SET_BIT, gpioNumber, (uint)Marshal.SizeOf(gpioNumber));
            else
                return SendIOCTL(IOCTL_GPIO_CLR_BIT, gpioNumber, (uint)Marshal.SizeOf(gpioNumber));
        }

        public bool SetAsInput(uint gpioNumber)
        {
            IoctlGpioInfo inputParameters = new IoctlGpioInfo(gpioNumber, GPIO_DIR_INPUT);

            return SendIOCTL(IOCTL_GPIO_SETMODE, inputParameters, (uint)Marshal.SizeOf(inputParameters));
        }

       
        /**
        //	Gets the current configuration for the given GPIO
        */
        public GpioConfiguration GetConfiguration(uint gpioNumber)
        {
            UInt32 result = GetIOCTL(IOCTL_GPIO_GETMODE, gpioNumber);
            if ((result & GPIO_DIR_INPUT)!=0)
                return GpioConfiguration.INPUT;
            else
                return GpioConfiguration.OUTPUT;

        }

        /**
        // Get current level of an output gpio
        */
        public GpioLevel GetGpioLevel(uint gpioNumber)
        {
            UInt32 uiRes = GetIOCTL(IOCTL_GPIO_GETBIT, gpioNumber);
            if (uiRes != 0)
                return GpioLevel.HIGH;
            else
                return GpioLevel.LOW;
        }

    }
}